"""
This module covers the raw output semantics of the OCPTV library.
"""
import dataclasses as dc
import json
import threading
import time
import typing as ty
from enum import Enum

from .config import Writer
from .objects import ArtifactType, Root, RootArtifactType, SchemaVersion

Primitive = ty.Union[float, int, bool, str, None]
JSON = ty.Union[ty.Dict[str, "JSON"], ty.List["JSON"], Primitive]


def _is_optional(field: ty.Type):
    # type hackery incoming
    # ty.Optional[T] == ty.Union[T, None]
    # since ty.Union[ty.Union[T,U]] = ty.Union[T,U] we can the
    # following transitiveness to check that type is optional
    return field == ty.Optional[field]


class ArtifactEmitter:
    """
    Serializes and emits the data on the configured output channel for the lib.
    Uses the low level dataclass models for the spec, but should not be used in user code.
    """

    def __init__(self, writer: Writer):
        self._seq_lock = threading.Lock()
        self._seq = 0

        self._writer = writer

        # use this event to ensure that the schema version message is the first to be
        # emitted, even when the initial writer was preempted
        self._version_emitted = threading.Event()

    @staticmethod
    def _serialize(artifact: ArtifactType):
        def visit(
            value: ty.Union[ArtifactType, ty.Dict, ty.List, Primitive],
            formatter: ty.Optional[ty.Callable[[ty.Any], str]] = None,
        ) -> JSON:
            # if present, formatter takes precedence over serialization
            if formatter is not None:
                return formatter(value)

            if dc.is_dataclass(value):
                obj: ty.Dict[str, JSON] = {}
                for field in dc.fields(value):
                    val = getattr(value, field.name)

                    if val is None:
                        if not _is_optional(field.type):
                            # TODO: fix exception text/type
                            raise RuntimeError("unacceptable none where not optional")

                        # for some reason, py3.7-3.10 fail to count the "continue" below as being covered
                        # by tests; this next line is a noop which avoids that behavior
                        val
                        continue

                    # spec_field takes precedence over spec_object
                    spec_field: ty.Optional[str] = field.metadata.get("spec_field", None)
                    spec_object: ty.Optional[str] = getattr(val, "SPEC_OBJECT", None)

                    if spec_field is None and spec_object is None:
                        # TODO: fix error type
                        raise RuntimeError(
                            "internal error, bad object decl: neither spec_field nor spec_object present for {}".format(
                                field.name
                            )
                        )

                    # safety: at least one is not None
                    name: str = ty.cast(str, spec_field or spec_object)
                    formatter = field.metadata.get("formatter", None)
                    obj[name] = visit(val, formatter)
                return obj
            elif isinstance(value, list) or isinstance(value, tuple):
                return [visit(k) for k in value]
            elif isinstance(value, dict):
                return {k: visit(v) for k, v in value.items()}
            elif isinstance(value, (str, float, int, bool, Enum)):
                return value

            raise RuntimeError("dont know how to serialize {}", value)

        return json.dumps(visit(artifact))

    def emit(self, artifact: RootArtifactType):
        """
        Emit the given artifact after serialization to JSON.
        This method is threadsafe.
        """
        seq = self._next_seq_no()
        if seq == 0:
            self._emit_version()
            seq = self._next_seq_no()

        # wait to make sure that version schema was written first
        self._version_emitted.wait()

        root = Root(
            impl=artifact,
            sequence_number=seq,
            timestamp=time.time(),
        )
        self._writer.write(self._serialize(root))

    def _emit_version(self):
        # use defaults for schema version here, should be set to latest
        root = Root(
            impl=SchemaVersion(),
            # seq number is always 0 for the spec version artifact
            sequence_number=0,
            timestamp=time.time(),
        )
        self._writer.write(self._serialize(root))

        # awake all threads that may have been waiting on version first write
        self._version_emitted.set()

    def _next_seq_no(self) -> int:
        with self._seq_lock:
            seq = self._seq
            self._seq += 1

        return seq



import inspect
import textwrap
import typing as ty

_RT = ty.TypeVar("_RT")


def export_api(ctor: ty.Callable[..., _RT]) -> ty.Callable[..., _RT]:
    """
    This decorator is functionally a no-op. It's just a marker to show lib maintainers
    that a specific method or class is/should be part of the public exports for the lib.
    The semantic of "public exports" here means that the user may create instances
    of these types directly. Most other public types are returned by lib code due to
    them containing private impl details.

    If applied to a class type, a suffix will be appended to its docstring to specify
    that it is a public export and can be directly instantiated by user code.

    All public export items must be re-exported in `__init__.py` at top module level.
    """

    if inspect.isclass(ctor):
        ctor.__doc__ = "{}\nThis type can be instantiated by user code directly.".format(
            textwrap.dedent(ctor.__doc__ or ""),
        )
    return ctor



import traceback
import typing as ty

from ocptv.api import export_api

from .objects import SourceLocation as SourceLocationSpec


@export_api
class SourceLocation:
    """
    The ``SourceLocation`` instances are coordinates in user diagnostic code, used
    for documenting where an ``Error``, ``Log`` or ``Diagnosis`` artifacts were emitted from.

    For functions taking a ``source_location`` parameter, eg. the ones making error messages,
    the default value is ``SourceLocation.CALLER``. Thi special instance is used to tell the
    library to try to guess the location in user code based on the call stack.

    However, the ``source_location`` can also be manually specified if needed.

    Usage:

    .. code-block:: python

        import ocptv.output as tv
        run = tv.TestRun(name="test", version="1.0")
        # default source_location=SourceLocation.CALLER on the following line
        # will result in the fields identifying the next line itself
        run.add_log(LogSeverity.DEBUG, "Some interesting message.")

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#sourcelocation
    """

    CALLER: "SourceLocation"
    """ Special instance. See ``SourceLocation`` docstring for details. """

    def __init__(self, filename: str, line_number: int):
        """
        :param filename: diagnostic package source filename being referenced.
        :param line_number: line number inside the diagnostic package source file.
        """
        self._file = filename
        self._line = line_number

    def to_spec(self) -> ty.Optional[SourceLocationSpec]:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        return SourceLocationSpec(file=self._file, line=self._line)


# note: the field values here dont matter. This is used as a guard value.
# Equality comparison should be done by instance.
SourceLocation.CALLER = SourceLocation("", 0)


class NullSourceLocation(SourceLocation):
    """
    This is an internal specialization of ``SourceLocation`` that will result in the
    ``source_location`` message fields not to be emitted by virtue of translating into a None.
    """

    def __init__(self):
        super().__init__("", 0)

    def to_spec(self) -> ty.Optional[SourceLocationSpec]:
        return None


def get_caller_source(offset: int = 1) -> SourceLocation:
    """
    Get the caller site coordinates as a SourceLocation instance.
    This will try to guess the callsite in user code by going down the stack.

    :param offset: expected number of stack frames to user code from the
        context of the code calling this function. Default value is 1, as it's
        expected that this function is called from lib code that is exactly one
        level from the user code.
    :returns: if the user code coordinates could be identified, this returns
        a valid SourceLocation object; otherwise it returns a null-value specialization that
        results in the JSON output to not be emitted.
    """
    assert offset > 0

    # TODO: this may be refactored later on to look at frame data, rather than use an index
    frames = traceback.extract_stack()[::-1]
    try:
        # the +1 is this function frame
        caller = frames[offset + 1]

        lineno = 0
        if caller.lineno is not None:
            lineno = caller.lineno

        return SourceLocation(
            filename=caller.filename,
            line_number=lineno,
        )
    except IndexError:
        # unlikely: error trying to get the user code frame, so return an
        # instance that will translate to None in the spec
        return NullSourceLocation()



import dataclasses as dc
import sys
import typing as ty

if ty.TYPE_CHECKING:  # pragma: no cover
    # mypy extension for py37
    from typing_extensions import Protocol
else:
    Protocol = object

if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import get_args, get_origin
else:  # pragma: no cover
    # following are very simple substitutes for their py3.8+ equivalents, but
    # they are sufficient for the usage here in `check_field_types`
    def get_origin(typ: ty.Type):
        if not hasattr(typ, "__origin__"):
            return None
        return typ.__origin__

    def get_args(typ: ty.Type):
        if not hasattr(typ, "__args__"):
            return None
        # note: in py37, the __args__ will contain typevar instances if the generic
        # type was not parametrized in the first place. Remove them because they
        # serve no purpose in the type checking here.
        return tuple(x for x in typ.__args__ if not isinstance(x, ty.TypeVar))


from .config import get_config


class Dataclass(Protocol):
    """
    Protocol type to describe all low level serializable objects in this file.
    """

    __dataclass_fields__: ty.ClassVar[dict]


Primitive = ty.Union[float, int, bool, str, None]
CheckedValue = ty.Union[ty.Dict[str, "CheckedValue"], ty.List["CheckedValue"], Dataclass, Primitive]


def ellipsis(value: ty.Any, max_length: int = 64) -> str:
    s = str(value)
    if len(s) > max_length:
        return s[:max_length] + "..."
    return s


class TypeCheckError(TypeError):
    def __init__(
        self,
        value: CheckedValue,
        expected: str,
        *,
        trace: ty.List[str],
    ):
        self._value = value
        self._expected = expected
        self._trace = trace

        parts = [
            "Cannot use type '{}'".format(type(value).__name__),
            "  Expected type '{}'".format(expected),
            "  Path to offending field is '{}'".format(" -> ".join(trace)),
        ]
        self._msg = "\n".join(parts)

    def __str__(self):
        return self._msg

    @property
    def value(self):
        return self._value

    @property
    def expected(self):
        return self._expected

    @property
    def trace(self):
        return self._trace


class UnionCheckError(TypeCheckError):
    def __init__(
        self,
        value: CheckedValue,
        expected: str,
        *,
        trace: ty.List[str],
        tries: ty.List[ty.Union[TypeCheckError, "UnionCheckError"]],
    ):
        super().__init__(value, expected, trace=trace)
        self._tries = tries

        parts = [
            "Cannot use type '{}' for field value".format(type(value).__name__),
            "  Expected type '{}'".format(expected),
            "  Path to offending field is '{}'".format(" -> ".join(trace)),
        ]
        parts.append("  Tried checks:")

        def format_tries(tries, indent=1):
            for t in tries:
                pre = "  " * indent
                parts.append("{}Value '{}' did not match expected type '{}'".format(pre, ellipsis(t.value), t.expected))
                parts.append("{}Path to field is '{}'".format(pre, " -> ".join(t.trace)))
                if type(t) is UnionCheckError:
                    format_tries(t._tries, indent + 1)

        format_tries(tries)
        self._msg = "\n".join(parts)

    def __str__(self):
        return self._msg


def _check_type_any(obj: CheckedValue, hint: ty.Type, trace: ty.List[str]):
    type_origin = get_origin(hint)
    type_args = get_args(hint)

    if type_origin is list:
        # generic type: typ == ty.List[...]
        if len(type_args) != 1:
            # for non-parametric generics, args is () so we can't actually
            # check for the list item types; short return
            return

        # for list, alias nparams=1
        item_type = type_args[0]

        if not isinstance(obj, list):
            raise TypeCheckError(obj, expected=f"List [{item_type}]", trace=trace)

        # unpack generic type to list item
        for i, v in enumerate(obj):
            subtrace = trace + [f"List[{i}]"]
            _check_type_any(v, item_type, subtrace)

    elif type_origin is dict:
        # generic type: typ == ty.Dict[...]
        if len(type_args) != 2:
            # for non-parametric generics, args is () so we can't actually
            # check for the dict keys or values types; short return
            return

        # for dict, alias nparams=2
        key_type, value_type = type_args

        if not isinstance(obj, dict):
            raise TypeCheckError(obj, expected=f"Dict [{key_type}, {value_type}]", trace=trace)

        # unpack generic type to kv tuple
        for k, v in obj.items():
            subtrace = trace + [f"Dict[{k}]"]
            _check_type_any(k, key_type, subtrace)
            _check_type_any(v, value_type, subtrace)

    elif type_origin is ty.Union:
        # for other generic types, get the covered types
        # note: ty.Optional is ty.Union[x, type(None)] so it still fits here

        any_passed = False
        errors = []
        for hint_item in type_args:
            try:
                _check_type_any(obj, hint_item, trace)
                any_passed = True
                break
            except TypeCheckError as e:
                errors.append(e)

        if not any_passed:
            raise UnionCheckError(obj, expected=f"Union[{type_args}]", trace=trace, tries=errors)

    elif type_origin is not None:
        raise ValueError(f"unsupported type origin value '{type_origin}'")

    elif hint is ty.Any:
        # we can't check anything if type is any, so just return
        return

    elif dc.is_dataclass(obj):
        for field in dc.fields(obj):
            subtrace = trace + [f"{obj.__class__.__name__}.{field.name}"]
            _check_type_any(getattr(obj, field.name), field.type, subtrace)

    elif not isinstance(obj, hint):
        raise TypeCheckError(obj, expected=hint.__name__, trace=trace)


def check_field_types(obj: Dataclass):
    """
    Check that the values inside the given dataclass' fields are of the correct
    type at runtime.
    This currently covers needed field types used in this lib, but is not generic
    enough to cover all typing configurations. See tests for more details.

    Can be disabled in lib config. See `ocptv.output.config()`.

    :throws TypeError on failures.
    """
    if not get_config().enable_runtime_checks:
        return

    _check_type_any(obj, type(obj), trace=[])



"""
This module contains output channel configuration for the OCPTV library.
"""
import threading
import typing as ty
from abc import ABC, abstractmethod
from datetime import timezone, tzinfo

from ocptv.api import export_api


class Writer(ABC):  # pragma: no cover
    """
    Abstract writer interface for the lib. Should be used as a base for
    any output writer implementation (for typing purposes).
    NOTE: Writer impls must ensure thread safety.
    """

    @abstractmethod
    def write(self, buffer: str):
        pass


@export_api
class StdoutWriter(Writer):
    """
    A simple writer that prints the json to stdout.
    """

    def __init__(self):
        self._lock = threading.Lock()

    def write(self, buffer: str):
        with self._lock:
            print(buffer)


class Config:
    """
    Thread safe storage for module configuration.

    Note: once a test run has started, configuration is considered committed and will
    continue to be used until a new test run is instantiated. Normally, there should
    just be one `TestRun` object per execution, hence this behavior is unlikely to be
    encountered.
    """

    def __init__(self):
        self._lock = threading.Lock()
        self._writer: Writer = StdoutWriter()
        self._enable_runtime_checks = True
        self._tzinfo: ty.Union[tzinfo, None] = timezone.utc

    @property
    def writer(self) -> Writer:
        with self._lock:
            return self._writer

    @writer.setter
    def writer(self, writer: Writer):
        with self._lock:
            self._writer = writer

    @property
    def enable_runtime_checks(self) -> bool:
        with self._lock:
            return self._enable_runtime_checks

    @enable_runtime_checks.setter
    def enable_runtime_checks(self, enable: bool):
        with self._lock:
            self._enable_runtime_checks = enable

    @property
    def timezone(self) -> ty.Union[tzinfo, None]:
        with self._lock:
            return self._tzinfo

    @timezone.setter
    def timezone(self, tzinfo: ty.Union[tzinfo, None]):
        with self._lock:
            self._tzinfo = tzinfo


# module scoped configuration (similar to python logging)
_config: Config = Config()

if ty.TYPE_CHECKING:  # pragma: no cover
    # We use a sentinel object that looks like a None for the static type checker.
    # At runtime, we check against the specific instance of `object()` to make a decision
    # on whether this parameter was passed in or not.
    # This is useful for types where NoneType is actually a valid instance of the type.
    _NOT_SET = None
else:
    _NOT_SET = object()


@export_api
def config(
    *,
    writer: ty.Optional[Writer] = None,
    enable_runtime_checks: ty.Optional[bool] = None,
    timezone: ty.Union[tzinfo, None] = _NOT_SET,
):
    """
    Configure how the ocptv.output lib behaves.

    :param writer: if provided, set the output channel writer.
    :param enable_runtime_checks: if provided, enables or disables runtime type checks.
    :param timezone: if provided, sets the timezone for the output formatted datetime fields.
        Use `None` to automatically determine the local system timezone.
        The library default, if never configured, is UTC.
    """
    global _config

    if writer is not None:
        _config.writer = writer

    if enable_runtime_checks is not None:
        _config.enable_runtime_checks = enable_runtime_checks

    if timezone is not _NOT_SET:
        _config.timezone = timezone


def get_config() -> Config:
    """
    Get the module configuration.
    Internal usage only.
    """
    global _config
    return _config



"""
This module describes the high level test run and related objects.
"""
import sys
import threading
import typing as ty
from contextlib import contextmanager

from ocptv.api import export_api

from .config import get_config
from .dut import Dut, SoftwareInfo
from .emit import ArtifactEmitter
from .objects import (
    Error,
    Log,
    LogSeverity,
    RunArtifact,
    RunEnd,
    RunStart,
    TestResult,
    TestStatus,
)
from .source import SourceLocation, get_caller_source
from .step import TestStep


@export_api
class TestRunError(RuntimeError):
    """
    The ``TestRunError`` can be raised with a context scope started by ``TestRun.scope()``.
    Any instance will be caught and handled by the context and will result in ending the
    whole run with an error outcome.
    """

    __slots__ = ("status", "result")

    # this may be further restricted re. params
    def __init__(self, *, status: TestStatus, result: TestResult):
        """
        :param status: outcome status for the diagnostic package as a whole.
        :param result: determine whether the validation passed for this test run.
        """
        self.status = status
        self.result = result


@export_api
class TestRun:
    """
    The ``TestRun`` object is a stateful interface for the diagnostic run.
    It presents various methods to interact with the run itself or to make child artifacts.

    Usage:

    .. code-block:: python

        import ocptv.output as tv
        run = tv.TestRun(name="test", version="1.0")
        with run.scope(dut=ocptv.Dut(id="dut0")):
            pass

    For other usages, see the the ``examples`` folder in the root of the project.

    All the methods in this class are threadsafe.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#test-run-artifacts
    """

    def __init__(
        self,
        name: str,
        version: str,
        *,
        command_line: ty.Optional[str] = None,
        parameters: ty.Optional[ty.Dict[str, ty.Union[str, int]]] = None,
    ):
        """
        Make a new stateful test run model object.

        :param name: name of this test run/diagnostic package.
        :param version: version string for the test run/diagnostic package.
        :param command_line: process command line that the diagnostic was started with.
        :param parameters: a flat dictionary of unspecified input parameters
            used in the diagnostic package.

        Parameters that have a discovery process are left to be specified in
        the ``start()`` or ``scope()`` calls.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#testrunstart
        """
        self._name = name
        self._version = version

        if command_line is not None:
            self._cmdline = command_line
        else:
            self._cmdline = self._get_cmdline()

        self._params = parameters or {}

        # once a test run has started, all semantically descendant artifacts
        # must use the same emitter without interruption
        self._emitter = ArtifactEmitter(writer=get_config().writer)

        self._step_lock = threading.Lock()
        self._step_id: int = 0

    def start(self, *, dut: Dut):
        """
        Emit the test run start artifact.

        :param dut: device-under-test information. The DUT info is considered to have
            finished discovery at the point of starting a test run.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#testrunstart
        """
        start = RunStart(
            name=self.name,
            version=self._version,
            command_line=self.command_line,
            parameters=self._params,
            dut_info=dut.to_spec(),
        )
        self._emitter.emit(RunArtifact(impl=start))

    def end(self, *, status: TestStatus, result: TestResult):
        """
        Emit the test run end artifact.

        :param status: outcome status for the diagnostic package as a whole.
        :param result: determine whether the validation passed for this test run.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#testrunend
        """
        end = RunEnd(
            status=status,
            result=result,
        )
        self._emitter.emit(RunArtifact(impl=end))

    @contextmanager
    def scope(self, *, dut: Dut):
        """
        Start a scoped test run.
        Emits the test run start artifact and, at the end of the scope, emits the run end artifact.

        The scope can also be exited sooner by raising the ``TestRunError`` error type.
        If no such exception was raised, the test run will end with status COMPLETE and result PASS.
        Any other exceptions are not handled and will pass through.

        Usage:

        .. code-block:: python

            import ocptv.output as tv
            run = tv.TestRun(...)
            with run.scope(dut=dut):
                pass

            run = tv.TestRun(...)
            with run.scope(dut=dut):
                raise TestRunError(status=TestStatus.SKIP, result=TestResult.NOT_APPLICABLE)

        :param dut: device-under-test information. See ``start`` method.
        """
        try:
            yield self.start(dut=dut)
        except TestRunError as re:
            self.end(status=re.status, result=re.result)
        else:
            self.end(status=TestStatus.COMPLETE, result=TestResult.PASS)

    def add_step(self, name: str) -> TestStep:
        """
        Create a new test step for this test run.
        The step is not started immediately, and should be used with the ``start()`` or ``scope()`` methods.

        :param name: name of the step to create
        :returns: reference to a model of the test step

        For additional details on parameters and step start, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#teststepstart
        """

        with self._step_lock:
            step_id = self._step_id
            self._step_id += 1

        step = TestStep(name, step_id=step_id, emitter=self._emitter)
        return step

    def add_log(
        self,
        severity: LogSeverity,
        message: str,
        source_location: ty.Optional[SourceLocation] = SourceLocation.CALLER,
    ):
        """
        Emit a log entry artifact relevant for this test run.

        :param severity: level of this log entry.
            See: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#severity
        :param message: free-form message for this log entry.
        :param source_location: source coordinates inside the diagnostic package
            where this artifact was produced. Default value ``SourceLocation.CALLER`` means that this
            function will automatically populate the spec field based on the caller frame.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#log
        """

        if source_location is SourceLocation.CALLER:
            source_location = get_caller_source()

        log = Log(
            severity=severity,
            message=message,
            source_location=source_location.to_spec() if source_location else None,
        )
        self._emitter.emit(RunArtifact(impl=log))

    def add_error(
        self,
        *,
        symptom: str,
        message: ty.Optional[str] = None,
        software_infos: ty.Optional[ty.List[SoftwareInfo]] = None,
        source_location: ty.Optional[SourceLocation] = SourceLocation.CALLER,
    ):
        """
        Emit an error artifact relevant for this test run.

        :param symptom: free-form description of the error symptom.
        :param message: free-form message associated with this error.
        :param software_infos: a list of `SoftwareInfo` references (as created
            in the DUT discovery process), that are relevant for this error. This can be used to specify
            a causal relationship between a software component and this error.
        :param source_location: source coordinates inside the diagnostic package
            where this artifact was produced. Default value ``SourceLocation.CALLER`` means that this
            function will automatically populate the spec field based on the caller frame.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#error
        """

        if software_infos is None:
            software_infos = []

        if source_location is SourceLocation.CALLER:
            source_location = get_caller_source()

        error = Error(
            symptom=symptom,
            message=message,
            software_infos=[o.to_spec() for o in software_infos],
            source_location=source_location.to_spec() if source_location else None,
        )
        self._emitter.emit(RunArtifact(impl=error))

    def _get_cmdline(self):
        return " ".join(sys.argv[1:])

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def command_line(self) -> str:
        return self._cmdline

    @property
    def parameters(self) -> ty.Optional[ty.Dict[str, ty.Union[str, int]]]:
        return self._params



import threading
import time
import typing as ty
from contextlib import contextmanager

from ocptv.api import export_api

from .dut import HardwareInfo, Subcomponent
from .emit import ArtifactEmitter
from .objects import (
    MeasurementSeriesElement,
    MeasurementSeriesEnd,
    MeasurementSeriesStart,
    MeasurementSeriesType,
    MeasurementValueType,
    Metadata,
    StepArtifact,
)
from .objects import Validator as ValidatorSpec
from .objects import ValidatorType, ValidatorValueType


class MeasurementSeriesEmitter(ArtifactEmitter):
    def __init__(self, step_id: str, emitter: ArtifactEmitter):
        self._step_id = step_id
        self._emitter = emitter

    def emit_impl(self, impl: MeasurementSeriesType):
        self._emitter.emit(StepArtifact(id=self._step_id, impl=impl))


# Following object is a proxy type so we get future flexibility, avoiding the usage
# of the low-level models.
@export_api
class Validator:
    """
    The ``Validator`` object represents a named validation that is relevant to a measurement or
    measurement series.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#validator
    """

    def __init__(
        self,
        *,
        type: ValidatorType,
        value: ValidatorValueType,
        name: ty.Optional[str] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        """
        Initialize a new validator object.

        :param type: classification for this validator.
            See: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#validatortype
        :param value: reference value for this validator. Can be a primitive type (int, float, str, bool)
            or a homogenous list of the same primitives. The list is only valid for the set-type validations.
        :param name: identification for this validator item.
        :param metadata: dictionary with unspecified metadata for this validator.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#validator
        """

        self._spec_object = ValidatorSpec(
            name=name,
            type=type,
            value=value,
            metadata=metadata,
        )

    def to_spec(self) -> ValidatorSpec:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        return self._spec_object


class MeasurementSeries:
    """
    The ``MeasurementSeries`` instances model a specific time-based list of values relevant to the diagnostic.
    A series is started by default on instantiation and must be ended with the ``.end()`` method or
    by using a ``.scope()`` context manager.

    Instances of this type must only be created by calls to ``TestStep.start_measurement_series()``.

    All the methods in this class are threadsafe.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurementseriesstart
    """

    def __init__(
        self,
        emitter: MeasurementSeriesEmitter,
        series_id: str,
        *,
        name: str,
        unit: ty.Optional[str] = None,
        validators: ty.Optional[ty.List[Validator]] = None,
        hardware_info: ty.Optional[HardwareInfo] = None,
        subcomponent: ty.Optional[Subcomponent] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        self._emitter = emitter
        self._id = series_id

        self._index_lock = threading.Lock()
        self._index: int = 0

        self._start(name, unit, validators, hardware_info, subcomponent, metadata)

    def add_measurement(
        self,
        *,
        value: MeasurementValueType,
        timestamp: ty.Optional[float] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        """
        Emit a new measurement item for this series.

        :param value: value of the taken measurement.
        :param timestamp: wallclock time when this measurement was taken. If unspecified,
            it will be computed based on the current wallclock time on the system running the diagnostic.
        :param metadata: dictionary with unspecified metadata for this measurement item.
        """

        if timestamp is None:
            # use local time if not specified
            timestamp = time.time()

        with self._index_lock:
            index = self._index
            self._index += 1

        measurement = MeasurementSeriesElement(
            index=index,
            value=value,
            timestamp=timestamp,
            series_id=self._id,
            metadata=metadata,
        )
        self._emitter.emit_impl(measurement)

    def _start(
        self,
        name: str,
        unit: ty.Optional[str] = None,
        validators: ty.Optional[ty.List[Validator]] = None,
        hardware_info: ty.Optional[HardwareInfo] = None,
        subcomponent: ty.Optional[Subcomponent] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        if validators is None:
            validators = []

        start = MeasurementSeriesStart(
            name=name,
            unit=unit,
            series_id=self._id,
            validators=[v.to_spec() for v in validators],
            hardware_info=hardware_info.to_spec() if hardware_info else None,
            subcomponent=subcomponent.to_spec() if subcomponent else None,
            metadata=metadata,
        )
        self._emitter.emit_impl(start)

    def end(self):
        """
        Emit a measurement series end artifact.

        Specification reference:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurementseriesend
        """

        end = MeasurementSeriesEnd(
            series_id=self._id,
            total_count=self._index,
        )
        self._emitter.emit_impl(end)

    @contextmanager
    def scope(self):
        """
        Wrap the measurement series into a scope that guarantees exception safety.
        When this scope ends, whether normally or from as exception, the end artifact is emitted.

        Usage:

        .. code-block:: python

            fan_speeds = step.start_measurement_series(name="fan0", ...)
            with fan_speeds.scope():
                fan_speeds.add_measurement(value=4200)
        """

        try:
            yield
        finally:
            self.end()



"""
Low level object models that are to be serialized as JSON.

NOT PUBLIC API, these are not intended to be used by client code
unless explicitly exported as public in ``__init__.py``.

Developer notes:
A field can either have ``metadata.spec_field`` set or ``field.SPEC_OBJECT`` set, not both.
If ``SPEC_OBJECT`` is set, this field is an union type and serialization should take the
value in ``SPEC_OBJECT`` as the serialized field name. Otherwise, the ``metadata.spec_field``
says what the serializer should use for field name.
In general, ``metadata.spec_field`` should only be present for primitive types.
"""
import dataclasses as dc
import typing as ty
from enum import Enum

from ocptv.api import export_api
from ocptv.formatter import format_enum, format_timestamp

if ty.TYPE_CHECKING:  # pragma: no cover
    # mypy extension for py37
    from typing_extensions import Protocol
else:
    Protocol = object

from .config import get_config
from .runtime_checks import check_field_types


def format_timestamp_with_tzinfo(ts: float) -> str:
    """Curry form with timezone from config"""
    return format_timestamp(ts, tz=get_config().timezone)


class ArtifactType(Protocol):
    """
    Protocol type to describe all low level serializable objects in this file.
    """

    __dataclass_fields__: ty.ClassVar[dict]


class OCPVersion(Enum):
    """
    OCP Test & Validation spec version consts.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#schemaversion
    schema: https://github.com/opencomputeproject/ocp-diag-core/output/$defs/schemaVersion
    """

    VERSION_2_0 = (2, 0)


@dc.dataclass
class SchemaVersion:
    """
    Low-level model for the `schemaVersion` spec object.
    Specifies the version that should be used to interpret following json outputs.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#schemaversion
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/root.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/output/$defs/schemaVersion
    """

    SPEC_OBJECT: ty.ClassVar[str] = "schemaVersion"

    major: int = dc.field(
        default=OCPVersion.VERSION_2_0.value[0],
        metadata={"spec_field": "major"},
    )

    minor: int = dc.field(
        default=OCPVersion.VERSION_2_0.value[1],
        metadata={"spec_field": "minor"},
    )

    def __post_init__(self):
        check_field_types(self)


# NOTE: This is intentionally not a dataclass. It is exported as public api
# because it does not have any inherent structure, other than a dict collection.
@export_api
class Metadata(dict):
    """
    Metadata container to accomodate any arbitrary keyed data put into it.
    For JSON output to work best, keys should be valid strings.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#metadata
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/metadata.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/metadata
    """

    SPEC_OBJECT: ty.ClassVar[str] = "metadata"


@dc.dataclass
class SourceLocation:
    """
    Provides information about which file/line of the source code in
    the diagnostic package generated the output.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#sourcelocation
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/source_location.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/sourceLocation
    """

    SPEC_OBJECT: ty.ClassVar[str] = "sourceLocation"

    file: str = dc.field(
        metadata={"spec_field": "file"},
    )

    line: int = dc.field(
        metadata={"spec_field": "line"},
    )


class LogSeverity(Enum):
    """
    Known log severity variants.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#severity
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/log.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/log/$defs/severity
    """

    INFO = "INFO"
    DEBUG = "DEBUG"
    WARNING = "WARNING"
    ERROR = "ERROR"
    FATAL = "FATAL"


@dc.dataclass
class Log:
    """
    Low-level model for `log` spec object.
    Is currently relevant for test run and test step artifact types.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#log
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/log.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/log
    """

    SPEC_OBJECT: ty.ClassVar[str] = "log"

    severity: LogSeverity = dc.field(
        metadata={
            "spec_field": "severity",
            "formatter": format_enum,
        },
    )

    message: str = dc.field(
        metadata={"spec_field": "message"},
    )

    source_location: ty.Optional[SourceLocation] = dc.field(
        metadata={"spec_field": "sourceLocation"},
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class File:
    """
    Low-level model for the `file` spec object.
    Represents a file artifact that was generated by running the diagnostic.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#file
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/file.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/file
    """

    SPEC_OBJECT: ty.ClassVar[str] = "file"

    name: str = dc.field(
        metadata={"spec_field": "displayName"},
    )

    uri: str = dc.field(
        metadata={"spec_field": "uri"},
    )

    is_snapshot: bool = dc.field(
        metadata={"spec_field": "isSnapshot"},
    )

    description: ty.Optional[str] = dc.field(
        metadata={"spec_field": "description"},
    )

    content_type: ty.Optional[str] = dc.field(
        metadata={"spec_field": "contentType"},
    )

    metadata: ty.Optional[Metadata]

    def __post_init__(self):
        check_field_types(self)


class SubcomponentType(Enum):
    """
    Type specification variants for DUT physical subcomponents.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#subcomponenttype
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/subcomponent.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/subcomponent/$defs/type
    """

    UNSPECIFIED = "UNSPECIFIED"
    ASIC = "ASIC"
    ASIC_SUBSYSTEM = "ASIC-SUBSYSTEM"
    BUS = "BUS"
    FUNCTION = "FUNCTION"
    CONNECTOR = "CONNECTOR"


@dc.dataclass
class Subcomponent:
    """
    Low-level model for the `subcomponent` spec object.
    Represents a physical subcomponent of a DUT hardware element.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#subcomponent
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/subcomponent.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/subcomponent
    """

    SPEC_OBJECT: ty.ClassVar[str] = "subcomponent"

    type: ty.Optional[SubcomponentType] = dc.field(
        metadata={
            "spec_field": "type",
            "formatter": format_enum,
        },
    )

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    location: ty.Optional[str] = dc.field(
        metadata={"spec_field": "location"},
    )

    version: ty.Optional[str] = dc.field(
        metadata={"spec_field": "version"},
    )

    revision: ty.Optional[str] = dc.field(
        metadata={"spec_field": "revision"},
    )

    def __post_init__(self):
        check_field_types(self)


class DiagnosisType(Enum):
    """
    Outcome of a diagnosis operation.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#diagnosistype
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/diagnosis.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/diagnosis/$defs/type
    """

    PASS = "PASS"
    FAIL = "FAIL"
    UNKNOWN = "UNKNOWN"


@dc.dataclass
class PlatformInfo:
    """
    Low-level model for the `platformInfo` spec object.
    Describe platform specific attributes of the DUT.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#platforminfo
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/dut_info.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/dutInfo/$defs/platformInfo
    """

    SPEC_OBJECT: ty.ClassVar[str] = "platformInfo"

    info: str = dc.field(
        metadata={"spec_field": "info"},
    )

    def __post_init__(self):
        check_field_types(self)


class SoftwareType(Enum):
    """
    Type specification for a software component of the DUT.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#softwaretype
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/dut_info.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/dutInfo/$defs/softwareInfo/properties/softwareType
    """

    UNSPECIFIED = "UNSPECIFIED"
    FIRMWARE = "FIRMWARE"
    SYSTEM = "SYSTEM"
    APPLICATION = "APPLICATION"


@dc.dataclass
class SoftwareInfo:
    """
    Low-level model for the `softwareInfo` spec object.
    Represents information of a discovered or exercised software component of the DUT.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#softwareinfo
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/dut_info.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/dutInfo/$defs/softwareInfo
    """

    SPEC_OBJECT: ty.ClassVar[str] = "softwareInfo"

    id: str = dc.field(
        metadata={"spec_field": "softwareInfoId"},
    )

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    version: ty.Optional[str] = dc.field(
        metadata={"spec_field": "version"},
    )

    revision: ty.Optional[str] = dc.field(
        metadata={"spec_field": "revision"},
    )

    type: ty.Optional[SoftwareType] = dc.field(
        metadata={
            "spec_field": "softwareType",
            "formatter": format_enum,
        },
    )

    computer_system: ty.Optional[str] = dc.field(
        metadata={"spec_field": "computerSystem"},
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class HardwareInfo:
    """
    Low-level model for the `hardwareInfo` spec object.
    Represents information of an enumerated or exercised hardware component of the DUT.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#hardwareinfo
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/dut_info.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/dutInfo/$defs/hardwareInfo
    """

    SPEC_OBJECT: ty.ClassVar[str] = "hardwareInfo"

    id: str = dc.field(
        metadata={"spec_field": "hardwareInfoId"},
    )

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    version: ty.Optional[str] = dc.field(
        metadata={"spec_field": "version"},
    )

    revision: ty.Optional[str] = dc.field(
        metadata={"spec_field": "revision"},
    )

    location: ty.Optional[str] = dc.field(
        metadata={"spec_field": "location"},
    )

    serial_no: ty.Optional[str] = dc.field(
        metadata={"spec_field": "serialNumber"},
    )

    part_no: ty.Optional[str] = dc.field(
        metadata={"spec_field": "partNumber"},
    )

    manufacturer: ty.Optional[str] = dc.field(
        metadata={"spec_field": "manufacturer"},
    )

    manufacturer_part_no: ty.Optional[str] = dc.field(
        metadata={"spec_field": "manufacturerPartNumber"},
    )

    odata_id: ty.Optional[str] = dc.field(
        metadata={"spec_field": "odataId"},
    )

    computer_system: ty.Optional[str] = dc.field(
        metadata={"spec_field": "computerSystem"},
    )

    manager: ty.Optional[str] = dc.field(
        metadata={"spec_field": "manager"},
    )

    def __post_init__(self):
        check_field_types(self)


def format_hardware_info(obj: HardwareInfo) -> str:
    return obj.id


@dc.dataclass
class DutInfo:
    """
    Low-level model for the `dutInfo` spec object.
    Contains all relevant information describing the DUT.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#dutinfo
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/dut_info.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/dutInfo
    """

    SPEC_OBJECT: ty.ClassVar[str] = "dutInfo"

    id: str = dc.field(
        metadata={"spec_field": "dutInfoId"},
    )

    name: ty.Optional[str] = dc.field(
        metadata={"spec_field": "name"},
    )

    platform_infos: ty.List[PlatformInfo] = dc.field(
        metadata={"spec_field": "platformInfos"},
    )

    software_infos: ty.List[SoftwareInfo] = dc.field(
        metadata={"spec_field": "softwareInfos"},
    )

    hardware_infos: ty.List[HardwareInfo] = dc.field(
        metadata={"spec_field": "hardwareInfos"},
    )

    metadata: ty.Optional[Metadata]

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class Diagnosis:
    """
    Low-level model for the `diagnosis` spec object.
    Contains the verdict given by the diagnostic regarding the DUT that was inspected.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#diagnosis
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/diagnosis.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/diagnosis
    """

    SPEC_OBJECT: ty.ClassVar[str] = "diagnosis"

    verdict: str = dc.field(
        metadata={"spec_field": "verdict"},
    )

    type: DiagnosisType = dc.field(
        metadata={
            "spec_field": "type",
            "formatter": format_enum,
        },
    )

    message: ty.Optional[str] = dc.field(
        metadata={"spec_field": "message"},
    )

    hardware_info: ty.Optional[HardwareInfo] = dc.field(
        metadata={
            "spec_field": "hardwareInfoId",
            "formatter": format_hardware_info,
        }
    )

    subcomponent: ty.Optional[Subcomponent]

    source_location: ty.Optional[SourceLocation] = dc.field(
        metadata={"spec_field": "sourceLocation"},
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class Error:
    """
    Low-level model for the `error` spec object.
    Represents an error encountered by the diagnostic software. It may refer to a DUT
    component or the diagnostic itself.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#error
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/error.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/error
    """

    SPEC_OBJECT: ty.ClassVar[str] = "error"

    symptom: str = dc.field(
        metadata={"spec_field": "symptom"},
    )

    message: ty.Optional[str] = dc.field(
        metadata={"spec_field": "message"},
    )

    software_infos: ty.List[SoftwareInfo] = dc.field(
        metadata={
            "spec_field": "softwareInfoIds",
            "formatter": lambda objs: [o.id for o in objs],
        },
    )

    source_location: ty.Optional[SourceLocation] = dc.field(
        metadata={"spec_field": "sourceLocation"},
    )

    def __post_init__(self):
        check_field_types(self)


MeasurementValueType = ty.Union[float, int, bool, str]

ValidatorValueType = ty.Union[
    ty.List[float],
    ty.List[int],
    ty.List[bool],
    ty.List[str],
    MeasurementValueType,
]


class ValidatorType(Enum):
    """
    Type specification for a measurement or series element validator logic.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#validator
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/validator.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/validator/$defs/type
    """

    EQUAL = "EQUAL"
    NOT_EQUAL = "NOT_EQUAL"
    LESS_THAN = "LESS_THAN"
    LESS_THAN_OR_EQUAL = "LESS_THAN_OR_EQUAL"
    GREATER_THAN = "GREATER_THAN"
    GREATER_THAN_OR_EQUAL = "GREATER_THAN_OR_EQUAL"
    REGEX_MATCH = "REGEX_MATCH"
    REGEX_NO_MATCH = "REGEX_NO_MATCH"
    IN_SET = "IN_SET"
    NOT_IN_SET = "NOT_IN_SET"


@dc.dataclass
class Validator:
    """
    Low-level model for the `validator` spec object.
    Contains the validation logic that the diagnostic applied for a specific measurement.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#validator
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/validator.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/validator
    """

    SPEC_OBJECT: ty.ClassVar[str] = "validator"

    name: ty.Optional[str] = dc.field(
        metadata={"spec_field": "name"},
    )

    type: ValidatorType = dc.field(
        metadata={
            "spec_field": "type",
            "formatter": format_enum,
        },
    )

    value: ValidatorValueType = dc.field(
        metadata={"spec_field": "value"},
    )

    metadata: ty.Optional[Metadata]

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class Measurement:
    """
    Low-level model for the `measurement` spec object.
    Represents an individual measurement taken by the diagnostic regarding the DUT.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurement
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/measurement.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/measurement
    """

    SPEC_OBJECT: ty.ClassVar[str] = "measurement"

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    value: MeasurementValueType = dc.field(
        metadata={"spec_field": "value"},
    )

    unit: ty.Optional[str] = dc.field(
        metadata={"spec_field": "unit"},
    )

    validators: ty.List[Validator] = dc.field(
        metadata={"spec_field": "validators"},
    )

    hardware_info: ty.Optional[HardwareInfo] = dc.field(
        metadata={
            "spec_field": "hardwareInfoId",
            "formatter": format_hardware_info,
        }
    )

    subcomponent: ty.Optional[Subcomponent]
    metadata: ty.Optional[Metadata]

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class MeasurementSeriesStart:
    """
    Low-level model for the `measurementSeriesStart` spec object.
    Start marker for a time based series of measurements.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurementseriesstart
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/measurement_series_start.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/measurementSeriesStart
    """

    SPEC_OBJECT: ty.ClassVar[str] = "measurementSeriesStart"

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    unit: ty.Optional[str] = dc.field(
        metadata={"spec_field": "unit"},
    )

    series_id: str = dc.field(
        metadata={"spec_field": "measurementSeriesId"},
    )

    validators: ty.List[Validator] = dc.field(
        metadata={"spec_field": "validators"},
    )

    hardware_info: ty.Optional[HardwareInfo] = dc.field(
        metadata={
            "spec_field": "hardwareInfoId",
            "formatter": format_hardware_info,
        }
    )

    subcomponent: ty.Optional[Subcomponent]
    metadata: ty.Optional[Metadata]

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class MeasurementSeriesEnd:
    """
    Low-level model for the `measurementSeriesEnd` spec object.
    End marker for a time based series of measurements.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurementseriesend
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/measurement_series_end.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/measurementSeriesEnd
    """

    SPEC_OBJECT: ty.ClassVar[str] = "measurementSeriesEnd"

    series_id: str = dc.field(
        metadata={"spec_field": "measurementSeriesId"},
    )

    total_count: int = dc.field(
        metadata={"spec_field": "totalCount"},
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class MeasurementSeriesElement:
    """
    Low-level model for the `measurementSeriesElement` spec object.
    Equivalent to the `Measurement` model but inside a time based series.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurementserieselement
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/measurement_series_element.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/measurementSeriesElement
    """

    SPEC_OBJECT: ty.ClassVar[str] = "measurementSeriesElement"

    index: int = dc.field(
        metadata={"spec_field": "index"},
    )

    value: MeasurementValueType = dc.field(
        metadata={"spec_field": "value"},
    )

    timestamp: float = dc.field(
        metadata={
            "spec_field": "timestamp",
            "formatter": format_timestamp_with_tzinfo,
        },
    )

    series_id: str = dc.field(
        metadata={"spec_field": "measurementSeriesId"},
    )

    metadata: ty.Optional[Metadata]

    def __post_init__(self):
        check_field_types(self)


MeasurementSeriesType = ty.Union[MeasurementSeriesStart, MeasurementSeriesEnd, MeasurementSeriesElement]

if ty.TYPE_CHECKING:
    # note: these must specify some bounds for the extension content in python, despite
    # the spec saying that it can be anything. This is necessary for the json serializer
    # to actually know how to output the data.
    ExtensionContentType = ty.Union[  # pragma: no cover
        ty.Dict[str, "ExtensionContentType"],
        ty.List["ExtensionContentType"],
        ty.Union[float, int, bool, str, None],
    ]
else:
    # the runtime checker cannot deal with recursive types, and this is meant to be any
    ExtensionContentType = ty.Any


@dc.dataclass
class Extension:
    """
    Low-level model for the `extension` spec object.
    Left as an implementation detail, the `Extension` just has a name and arbitrary data.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#extension
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_step_artifact.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testStepArtifact/$defs/extension
    """

    SPEC_OBJECT: ty.ClassVar[str] = "extension"

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    content: ExtensionContentType = dc.field(
        metadata={"spec_field": "content"},
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class RunStart:
    """
    Low-level model for the `testRunStart` spec object.
    Start marker for the beginning of a diagnostic test.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#testrunstart
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_run_start.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testRunStart
    """

    SPEC_OBJECT: ty.ClassVar[str] = "testRunStart"

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    version: str = dc.field(
        metadata={"spec_field": "version"},
    )

    command_line: str = dc.field(
        metadata={"spec_field": "commandLine"},
    )

    parameters: ty.Dict[str, ty.Any] = dc.field(
        metadata={"spec_field": "parameters"},
    )

    dut_info: DutInfo

    def __post_init__(self):
        check_field_types(self)


class TestStatus(Enum):
    """
    Represents the final execution status of a test.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#teststatus
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_status.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testStatus
    """

    # pytest incorrectly identifies this as a test
    __test__ = False  # type: ignore

    COMPLETE = "COMPLETE"
    ERROR = "ERROR"
    SKIP = "SKIP"


class TestResult(Enum):
    """
    Represents the final outcome of a test execution.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#testresult
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_run_end.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testRunEnd/$defs/testResult
    """

    # pytest incorrectly identifies this as a test
    __test__ = False  # type: ignore

    PASS = "PASS"
    FAIL = "FAIL"
    NOT_APPLICABLE = "NOT_APPLICABLE"


@dc.dataclass
class RunEnd:
    """
    Low-level model for the `testRunEnd` spec object.
    End marker signaling the finality of a diagnostic test.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#testrunend
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_run_end.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testRunEnd
    """

    SPEC_OBJECT: ty.ClassVar[str] = "testRunEnd"

    status: TestStatus = dc.field(
        metadata={
            "spec_field": "status",
            "formatter": format_enum,
        },
    )

    result: TestResult = dc.field(
        metadata={
            "spec_field": "result",
            "formatter": format_enum,
        },
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class RunArtifact:
    """
    Low-level model for the `testRunArtifact` spec object.
    Container for the run level artifacts.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#test-run-artifacts
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_run_artifact.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testRunArtifact
    """

    SPEC_OBJECT: ty.ClassVar[str] = "testRunArtifact"

    impl: ty.Union[RunStart, RunEnd, Log, Error]


@dc.dataclass
class StepStart:
    """
    Low-level model for the `testStepStart` spec object.
    Start marker for a test step inside a diagnosis run.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#teststepstart
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_step_start.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testStepStart
    """

    SPEC_OBJECT: ty.ClassVar[str] = "testStepStart"

    name: str = dc.field(
        metadata={"spec_field": "name"},
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class StepEnd:
    """
    Low-level model for the `testStepEnd` spec object.
    End marker for a test step inside a diagnosis run.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#teststepend
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_step_end.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testStepEnd
    """

    SPEC_OBJECT: ty.ClassVar[str] = "testStepEnd"

    status: TestStatus = dc.field(
        metadata={
            "spec_field": "status",
            "formatter": format_enum,
        },
    )

    def __post_init__(self):
        check_field_types(self)


@dc.dataclass
class StepArtifact:
    """
    Low-level model for the `testStepArtifact` spec object.
    Container for the step level artifacts.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#test-step-artifacts
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/test_step_artifact.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/testStepArtifact
    """

    SPEC_OBJECT: ty.ClassVar[str] = "testStepArtifact"

    id: str = dc.field(
        metadata={"spec_field": "testStepId"},
    )

    impl: ty.Union[
        StepStart,
        StepEnd,
        Measurement,
        MeasurementSeriesType,
        Diagnosis,
        Log,
        Error,
        File,
        Extension,
    ]

    def __post_init__(self):
        check_field_types(self)


RootArtifactType = ty.Union[SchemaVersion, RunArtifact, StepArtifact]


@dc.dataclass
class Root:
    """
    Low-level model for the root json node.
    Container for the root level artifacts.

    ref: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#outputartifacts
    schema url: https://github.com/opencomputeproject/ocp-diag-core/blob/main/json_spec/output/root.json
    schema ref: https://github.com/opencomputeproject/ocp-diag-core/output
    """

    impl: RootArtifactType

    sequence_number: int = dc.field(
        metadata={"spec_field": "sequenceNumber"},
    )

    timestamp: float = dc.field(
        metadata={
            "spec_field": "timestamp",
            "formatter": format_timestamp_with_tzinfo,
        },
    )

    def __post_init__(self):
        check_field_types(self)



from datetime import datetime, timedelta, timezone
from enum import Enum

from ocptv.formatter import format_enum, format_timestamp


def test_format_enum_simple():
    class TestEnum(Enum):
        V42 = 42

    assert format_enum(TestEnum.V42) == "42"


def test_format_timestamp_simple():
    tz = timezone(offset=timedelta(hours=1), name="plus1")
    value = datetime(
        year=2020,
        month=1,
        day=2,
        hour=3,
        minute=4,
        second=5,
        microsecond=100,
        tzinfo=tz,
    )
    assert format_timestamp(value.timestamp(), tz=tz) == "2020-01-02T03:04:05.000100+01:00"


def test_format_timestamp_zulu():
    assert format_timestamp(0, tz=timezone.utc) == "1970-01-01T00:00:00Z"



"""
This module describes the test steps inside the test run.
"""
import threading
import typing as ty
from contextlib import contextmanager

from ocptv.api import export_api

from .dut import HardwareInfo, SoftwareInfo, Subcomponent
from .emit import ArtifactEmitter
from .measurement import MeasurementSeries, MeasurementSeriesEmitter, Validator
from .objects import (
    Diagnosis,
    DiagnosisType,
    Error,
    Extension,
    ExtensionContentType,
    File,
    Log,
    LogSeverity,
    Measurement,
    MeasurementValueType,
    Metadata,
    StepArtifact,
    StepEnd,
    StepStart,
    TestStatus,
)
from .source import SourceLocation, get_caller_source


@export_api
class TestStepError(RuntimeError):
    """
    The ``TestStepError`` can be raised within a context scope started by ``TestStep.scope()``.
    Any instance will be caught and handled by the context and will result in ending the
    step with an error.
    """

    __slots__ = "status"

    def __init__(self, *, status: TestStatus):
        """
        :param status: outcome status for this step.
        """
        self.status = status


class TestStep:
    """
    The ``TestStep`` instances are the stateful interface to diagnostic steps inside a run.
    They present methods to interact with the steps themselves or to make child artifacts.

    Should not be used directly by user code, but created through ``TestRun.add_step()``.

    Usage:

    .. code-block:: python

        step0 = run.add_step("step0") # run: TestRun
        with step0.scope():
            pass

    For other usages, see the the ``examples`` folder in the root of the project.

    All the methods in this class are threadsafe.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#test-step-artifacts
    """

    def __init__(self, name: str, *, step_id: int, emitter: ArtifactEmitter):
        self._name = name
        self._id = step_id
        self._idstr = "{}".format(step_id)
        self._emitter = emitter

        # TODO: do we want manually controlled values for the series id?
        self._series_lock = threading.Lock()
        self._measurement_series_id: int = 0

    def start(self):
        """
        Emit a test step start artifact.

        Specification reference:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#teststepstart
        """

        start = StepStart(name=self._name)
        self._emitter.emit(StepArtifact(id=self._idstr, impl=start))

    def end(self, *, status: TestStatus):
        """
        Emit a test step end artifact.

        :param status: outcome status for this step.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#teststepend
        """

        end = StepEnd(status=status)
        self._emitter.emit(StepArtifact(id=self._idstr, impl=end))

    @contextmanager
    def scope(self):
        """
        Start a scoped test step.
        Emits the test step start artifact and, at the end of the scope, emits the step end artifact.

        The scope can also be exited sooner by raising the ``TestStepError`` error type.
        If no such exception was raised, the test step will end with status COMPLETE.
        Any other exceptions are not handled and will pass through.

        Usage:

        .. code-block:: python

            step0 = run.add_step(name="step0")
            with step0.scope():
                pass

            step0 = run.add_step(name="step0")
            with step0.scope():
                raise TestStepError(status=TestStatus.SKIP)
        """

        try:
            yield self.start()
        except TestStepError as e:
            self.end(status=e.status)
        else:
            self.end(status=TestStatus.COMPLETE)

    def add_measurement(
        self,
        *,
        name: str,
        value: MeasurementValueType,
        unit: ty.Optional[str] = None,
        validators: ty.Optional[ty.List[Validator]] = None,
        hardware_info: ty.Optional[HardwareInfo] = None,
        subcomponent: ty.Optional[Subcomponent] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        """
        Emit a single measurement artifact as taken by the test step.

        :param name: identification for this measurement item.
        :param value: value of the taken measurement.
        :param unit: free-form unit specification for this measurement.
        :param validator: specifies how this measurement was validated in the
            test step by the diagnostic package.
        :param hardware_info: reference to the hardware component that this
            measurement was taken from or is relative to.
        :param subcomponent: reference to the hardware subcomponent (lower than FRU
            level) that this measurement was taken from or is relative to.
        :param metadata: dictionary with unspecified metadata for this measurement item.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurement
        """

        if validators is None:
            validators = []

        measurement = Measurement(
            name=name,
            value=value,
            unit=unit,
            validators=[v.to_spec() for v in validators],
            hardware_info=hardware_info.to_spec() if hardware_info else None,
            subcomponent=subcomponent.to_spec() if subcomponent else None,
            metadata=metadata,
        )
        self._emitter.emit(StepArtifact(id=self._idstr, impl=measurement))

    # TODO: we start a measurement series implicitly but not test steps
    def start_measurement_series(
        self,
        *,
        name: str,
        unit: ty.Optional[str] = None,
        validators: ty.Optional[ty.List[Validator]] = None,
        hardware_info: ty.Optional[HardwareInfo] = None,
        subcomponent: ty.Optional[Subcomponent] = None,
        metadata: ty.Optional[Metadata] = None,
    ) -> MeasurementSeries:
        """
        Start a new series of measurement and emit the series start artifact.

        :param name: identification for this measurement series.
        :param unit: free-form unit specification for this measurement series.
        :param validator: specifies how the measurement elements in this series
            will be validated in this test step by the diagnostic package.
        :param hardware_info: reference to the hardware component that this
            measurement series elements will be taken from or will be relative to.
        :param subcomponent: reference to the hardware subcomponent (lower than FRU
            level) that this measurement series elements will be taken from or will be relative to.
        :param metadata: dictionary with unspecified metadata for this measurement series.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#measurementseriesstart
        """

        with self._series_lock:
            series_id = self._measurement_series_id
            self._measurement_series_id += 1

        # TODO: arbitrary derivation for measurement id here, but unique for run scope;
        # may be a fea request that this follows a template
        series = MeasurementSeries(
            MeasurementSeriesEmitter(self._idstr, self._emitter),
            "{}_{}".format(self._id, series_id),
            name=name,
            unit=unit,
            validators=validators,
            hardware_info=hardware_info,
            subcomponent=subcomponent,
            metadata=metadata,
        )

        return series

    def add_diagnosis(
        self,
        diagnosis_type: DiagnosisType,
        *,
        verdict: str,
        message: ty.Optional[str] = None,
        hardware_info: ty.Optional[HardwareInfo] = None,
        subcomponent: ty.Optional[Subcomponent] = None,
        source_location: ty.Optional[SourceLocation] = SourceLocation.CALLER,
    ):
        """
        Emit a new diagnosis artifact as determined by this test step.

        :param diagnosis_type: outcome type, eg. pass/fail/unknown
            see https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#diagnosistype
        :param verdict: free-form specification of diagnosis outcome
        :param message: message associated with this diagnosis
        :param hardware_info: reference to a part of the DUT that was diagnosed
        :param subcomponent: reference to a subcomponent of the DUT hardware
        :param source_location: source coordinates inside the diagnostic package
            where this artifact was produced. Default value ``SourceLocation.CALLER`` means that this
            function will automatically populate the spec field based on the caller frame.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#diagnosis
        """
        if source_location is SourceLocation.CALLER:
            source_location = get_caller_source()

        diag = Diagnosis(
            verdict=verdict,
            type=diagnosis_type,
            message=message,
            hardware_info=hardware_info.to_spec() if hardware_info else None,
            subcomponent=subcomponent.to_spec() if subcomponent else None,
            source_location=source_location.to_spec() if source_location else None,
        )
        self._emitter.emit(StepArtifact(id=self._idstr, impl=diag))

    def add_log(
        self,
        severity: LogSeverity,
        message: str,
        source_location: ty.Optional[SourceLocation] = SourceLocation.CALLER,
    ):
        """
        Emit a log entry artifact relevant for this test step.

        :param severity: level of this log entry.
            See: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#severity
        :param message: free-form message for this log entry.
        :param source_location: source coordinates inside the diagnostic package
            where this artifact was produced. Default value ``SourceLocation.CALLER`` means that this
            function will automatically populate the spec field based on the caller frame.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#log
        """

        if source_location is SourceLocation.CALLER:
            source_location = get_caller_source()

        log = Log(
            severity=severity,
            message=message,
            source_location=source_location.to_spec() if source_location else None,
        )
        self._emitter.emit(StepArtifact(id=self._idstr, impl=log))

    def add_error(
        self,
        *,
        symptom: str,
        message: ty.Optional[str] = None,
        software_infos: ty.Optional[ty.List[SoftwareInfo]] = None,
        source_location: ty.Optional[SourceLocation] = SourceLocation.CALLER,
    ):
        """
        Emit an error artifact relevant for this test step.

        :param symptom: free-form description of the error symptom.
        :param message: free-form message associated with this error.
        :param software_infos: a list of `SoftwareInfo` references (as created
            in the DUT discovery process), that are relevant for this error. This can be used to specify
            a causal relationship between a software component and this error.
        :param source_location: source coordinates inside the diagnostic package
            where this artifact was produced. Default value ``SourceLocation.CALLER`` means that this
            function will automatically populate the spec field based on the caller frame.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#error
        """

        if software_infos is None:
            software_infos = []

        if source_location is SourceLocation.CALLER:
            source_location = get_caller_source()

        error = Error(
            symptom=symptom,
            message=message,
            software_infos=[o.to_spec() for o in software_infos],
            source_location=source_location.to_spec() if source_location else None,
        )
        self._emitter.emit(StepArtifact(id=self._idstr, impl=error))

    def add_file(
        self,
        *,
        name: str,
        uri: str,
        is_snapshot: bool = True,
        description: ty.Optional[str] = None,
        content_type: ty.Optional[str] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        """
        Emit an artifact specifying a file resource produced in the diagnosis process.

        :param name: identifying name for this file.
        :param uri: a local/remote location specification for this file.
        :param bool: specifies whether this file is a complete production or if it is a
            temporary snapshot as seen in the diagnosis process.
        :param description: free-form description for this file resource.
        :param content_type: MIME-type specification for this file contents.
        :param metadata: dictionary with unspecified metadata for this file.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#file
        """

        file = File(
            name=name,
            uri=uri,
            is_snapshot=is_snapshot,
            description=description,
            content_type=content_type,
            metadata=metadata,
        )
        self._emitter.emit(StepArtifact(id=self._idstr, impl=file))

    def add_extension(
        self,
        *,
        name: str,
        content: ExtensionContentType,
    ):
        """
        Emits a step extension artifact.
        Extensions are meant to accommodate any aspects that are not standardized and may be vendor specific.

        :param name: identification of this specific extension artifact
        :param content: any vendor specific content that can be serialized to JSON

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#extension
        """

        ext = Extension(name=name, content=content)
        self._emitter.emit(StepArtifact(id=self._idstr, impl=ext))

    @property
    def name(self) -> str:
        return self._name

    @property
    def id(self) -> int:
        return self._id



# following are the only public api exports
from .config import StdoutWriter, Writer, config
from .dut import Dut, Subcomponent
from .measurement import Validator
from .objects import (
    DiagnosisType,
    LogSeverity,
    Metadata,
    OCPVersion,
    SoftwareType,
    SubcomponentType,
    TestResult,
    TestStatus,
    ValidatorType,
)
from .run import TestRun, TestRunError
from .source import SourceLocation
from .step import TestStep, TestStepError

# the specification version that this module implements
# ref: https://github.com/opencomputeproject/ocp-diag-core/tree/5708d0c/json_spec
__spec_version__ = OCPVersion.VERSION_2_0



import threading
import typing as ty

from ocptv.api import export_api

from .objects import DutInfo
from .objects import HardwareInfo as HardwareInfoSpec
from .objects import Metadata
from .objects import PlatformInfo as PlatformInfoSpec
from .objects import SoftwareInfo as SoftwareInfoSpec
from .objects import SoftwareType
from .objects import Subcomponent as SubcomponentSpec
from .objects import SubcomponentType


class PlatformInfo:
    """
    Platform information for the DUT.

    This object should not be instantiated directly by user code.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#platforminfo
    """

    def __init__(self, info_tag: str):
        self._spec_object = PlatformInfoSpec(info=info_tag)

    def to_spec(self) -> PlatformInfoSpec:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        return self._spec_object


class SoftwareInfo:
    """
    Software information for a component of the DUT.

    This object should not be instantiated directly by user code.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#softwareinfo
    """

    def __init__(
        self,
        id: str,
        name: str,
        type: ty.Optional[SoftwareType] = None,
        version: ty.Optional[str] = None,
        revision: ty.Optional[str] = None,
        computer_system: ty.Optional[str] = None,
    ):
        self._spec_object = SoftwareInfoSpec(
            id=id,
            name=name,
            version=version,
            revision=revision,
            type=type,
            computer_system=computer_system,
        )

    def to_spec(self) -> SoftwareInfoSpec:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        return self._spec_object


class HardwareInfo:
    """
    Hardware information for a component of the DUT.

    This object should not be instantiated directly by user code.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#hardwareinfo
    """

    def __init__(
        self,
        id: str,
        name: str,
        version: ty.Optional[str] = None,
        revision: ty.Optional[str] = None,
        location: ty.Optional[str] = None,
        serial_no: ty.Optional[str] = None,
        part_no: ty.Optional[str] = None,
        manufacturer: ty.Optional[str] = None,
        manufacturer_part_no: ty.Optional[str] = None,
        odata_id: ty.Optional[str] = None,
        computer_system: ty.Optional[str] = None,
        manager: ty.Optional[str] = None,
    ):
        self._spec_object = HardwareInfoSpec(
            id=id,
            name=name,
            version=version,
            revision=revision,
            location=location,
            serial_no=serial_no,
            part_no=part_no,
            manufacturer=manufacturer,
            manufacturer_part_no=manufacturer_part_no,
            odata_id=odata_id,
            computer_system=computer_system,
            manager=manager,
        )

    def to_spec(self) -> HardwareInfoSpec:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        return self._spec_object


@export_api
class Dut:
    """
    The `Dut` instances model the specific devices under test used in this diagnostic package.

    All the methods in this class are threadsafe.

    Specification reference:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#dutinfo
    """

    def __init__(
        self,
        id: str,
        name: ty.Optional[str] = None,
        metadata: ty.Optional[Metadata] = None,
    ):
        """
        Initialize a new Dut instance.

        :param id: unique identifier for this device under test.
        :param name: any domain specific naming for the device.
        :param metadata: dictionary with unspecified metadata for the device.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#dutinfo
        """

        self._id = id
        self._name = name
        self._metadata = metadata

        self._info_lock = threading.Lock()
        self._platform_infos: ty.List[PlatformInfo] = []
        self._software_infos: ty.List[SoftwareInfo] = []
        self._hardware_infos: ty.List[HardwareInfo] = []

    def add_platform_info(self, info_tag: str) -> PlatformInfo:
        """
        Add a new platform information item to this DUT.

        :param info_tag: free-form information about the platform.
        :returns: model reference to the new platform info item.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#platforminfo
        """

        info = PlatformInfo(info_tag=info_tag)
        with self._info_lock:
            self._platform_infos.append(info)
        return info

    def add_software_info(
        self,
        name: str,
        type: ty.Optional[SoftwareType] = None,
        version: ty.Optional[str] = None,
        revision: ty.Optional[str] = None,
        computer_system: ty.Optional[str] = None,
    ) -> SoftwareInfo:
        """
        Add a new information item describing a software component for this DUT.

        :param name: software component name.
        :param type: classification of the software component type.
            see: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#softwaretype
        :param version: version information for the software.
        :param str: revision information for the software.
        :param computer_system: Redfish-type name of the computer system where
            this software component is executing.
        :returns: model reference to the new software component.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#softwareinfo
        """

        with self._info_lock:
            # reserve an index
            software_id = "{}_{}".format(self._id, len(self._software_infos))

            # TODO(mimir-d): arbitrary id derivation; does this need to be more readable?
            info = SoftwareInfo(
                id=software_id,
                name=name,
                type=type,
                version=version,
                revision=revision,
                computer_system=computer_system,
            )
            self._software_infos.append(info)
            return info

    def add_hardware_info(
        self,
        name: str,
        version: ty.Optional[str] = None,
        revision: ty.Optional[str] = None,
        location: ty.Optional[str] = None,
        serial_no: ty.Optional[str] = None,
        part_no: ty.Optional[str] = None,
        manufacturer: ty.Optional[str] = None,
        manufacturer_part_no: ty.Optional[str] = None,
        odata_id: ty.Optional[str] = None,
        computer_system: ty.Optional[str] = None,
        manager: ty.Optional[str] = None,
    ) -> HardwareInfo:
        """
        Add a new information item describing a hardware component of the DUT.

        :param name: hardware component name.
        :param version: version of the hardware component.
        :param revision: revision of the hardware component.
        :param location: unspecified representation of the hardware location.
        :param serial_no: hardware component serial number.
        :param part_no: hardware component part number.
        :param manufacturer: manufacturer name of the hardware component.
        :param manufacturer_part_no: hardware component part number as provided by manufacturer.
        :param odata_id: Redfish-type identification for this hardware component.
        :param computer_system: Redfish-type name of the computer system to which this
            hardware component is visible.
        :param manager: Redfish-type name of the manager of this hardware component. Typically
            this is an out-of-band device/system.
        :returns: model reference to the new hardware component.

        For additional details on parameters, see:
        - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#hardwareinfo
        """

        with self._info_lock:
            # reserve an index
            hardware_id = "{}_{}".format(self._id, len(self._hardware_infos))

            info = HardwareInfo(
                id=hardware_id,
                name=name,
                version=version,
                revision=revision,
                location=location,
                serial_no=serial_no,
                part_no=part_no,
                manufacturer=manufacturer,
                manufacturer_part_no=manufacturer_part_no,
                odata_id=odata_id,
                computer_system=computer_system,
                manager=manager,
            )

            self._hardware_infos.append(info)
            return info

    def to_spec(self) -> DutInfo:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        with self._info_lock:
            platforms = [x.to_spec() for x in self._platform_infos]
            software = [x.to_spec() for x in self._software_infos]
            hardware = [x.to_spec() for x in self._hardware_infos]

        return DutInfo(
            id=self._id,
            name=self._name,
            platform_infos=platforms,
            software_infos=software,
            hardware_infos=hardware,
            metadata=self._metadata,
        )


# Following object is a proxy type so we get future flexibility, avoiding the usage
# of the low-level models.
@export_api
class Subcomponent:
    """
    A lower-than-FRU (field replaceable unit) hardware item inside the DUT.

    For additional details on parameters, see:
    - https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#subcomponent
    """

    def __init__(
        self,
        *,
        name: str,
        type: ty.Optional[SubcomponentType] = None,
        location: ty.Optional[str] = None,
        version: ty.Optional[str] = None,
        revision: ty.Optional[str] = None,
    ):
        """
        Initialize a hardware subcomponent.

        :param name: hardware subcomponent name.
        :param type: classification of the subcomponent type.
            See: https://github.com/opencomputeproject/ocp-diag-core/tree/main/json_spec#subcomponenttype
        :param location: unspecified representation of the subcomponent location.
        :param version: version of the hardware subcomponent.
        :param revision: revision of the hardware subcomponent.
        """

        self._spec_object = SubcomponentSpec(
            type=type,
            name=name,
            location=location,
            version=version,
            revision=revision,
        )

    def to_spec(self) -> SubcomponentSpec:
        """
        Internal usage. Convert to low-level model.

        :meta private:
        """
        return self._spec_object



import typing as ty
from datetime import datetime, timezone, tzinfo
from enum import Enum


def format_enum(variant: Enum) -> str:
    """Format an enum by variant value"""
    return "{}".format(variant.value)


def format_timestamp(ts: float, tz: ty.Optional[tzinfo] = timezone.utc) -> str:
    """
    Format an unix timestamp in local tz.
    If the timezone coincides with UTC, returns zulu time format.
    """
    dt = datetime.fromtimestamp(ts)
    isostr = dt.astimezone(tz).isoformat()

    utcsuffix = "+00:00"
    if isostr.endswith(utcsuffix):
        return isostr[: -len(utcsuffix)] + "Z"
    return isostr
